import { pluralize } from 'inflection';
import { constants } from 'http2';
import { IncomingMessage, ServerResponse } from 'http';
import * as v from 'valibot';
import Negotiator from 'negotiator';
import {SerializerPair} from './serializers';

export interface DataSource<T = object> {
	initialize(): Promise<unknown>;
	getMultiple(): Promise<T[]>;
	getSingle(id: string): Promise<T | null>;
	create(data: Partial<T>): Promise<T>;
	delete(id: string): Promise<unknown>;
	emplace(id: string, data: Partial<T>): Promise<T>;
	patch(id: string, data: Partial<T>): Promise<T | null>;
}

export interface ApplicationParams {
	name: string;
	dataSource?: (resource: Resource) => DataSource;
}

export interface Resource {
	idAttr: string;
	itemName?: string;
	collectionName?: string;
	routeName?: string;
	dataSource: DataSource;
	newId(dataSource: DataSource): string | number | unknown;
}

interface GenerationStrategy {
	(dataSource: DataSource, ...args: unknown[]): Promise<string | number | unknown>;
}

interface IdParams {
	generationStrategy: GenerationStrategy;
}

export const resource = (schema: Parameters<typeof v.object>[0]) => {
	let theIdAttr: string;
	let theItemName: string;
	let theCollectionName: string;
	let theRouteName: string;
	let idGenerationStrategy: GenerationStrategy;
	const fullTextAttrs = new Set<string>();

	return {
		id(newIdAttr: string, params: IdParams) {
			theIdAttr = newIdAttr;
			idGenerationStrategy = params.generationStrategy;
			return this;
		},
		newId(dataSource: DataSource) {
			return idGenerationStrategy(dataSource);
		},
		fullText(fullTextAttr: string) {
			if (schema[fullTextAttr]?.type === 'string') {
				fullTextAttrs.add(fullTextAttr);
				return this;
			}

			throw new Error(`Could not set attribute ${fullTextAttr} as fulltext.`);
		},
		name(n: string) {
			theItemName = n;
			theCollectionName = theCollectionName ?? pluralize(theItemName).toLowerCase();
			theRouteName = theRouteName ?? theCollectionName;
			return this;
		},
		collection(n: string) {
			theCollectionName = n;
			theRouteName = theRouteName ?? theCollectionName;
			return this;
		},
		route(n: string) {
			theRouteName = n;
			return this;
		},
		get idAttr() {
			return theIdAttr;
		},
		get collectionName() {
			return theCollectionName;
		},
		get itemName() {
			return theItemName;
		},
		get routeName() {
			return theRouteName;
		}
	};
};

interface CreateServerParams {
	baseUrl?: string;
	host?: string;
}

const handleGetAll = async (serializerPair: SerializerPair, mediaType: string, dataSource: DataSource, res: ServerResponse) => {
	const resData = await dataSource.getMultiple(); // TODO paginated responses per resource
	const theFormatted = serializerPair.serialize(resData);

	res.writeHead(constants.HTTP_STATUS_OK, {
		'Content-Type': mediaType,
		'X-Resource-Total-Item-Count': resData.length
	});
	res.end(theFormatted);
};

const handleGetSingle = async (serializerPair: SerializerPair, mediaType: string, resource: Resource, mainResourceId: string, dataSource: DataSource, res: ServerResponse) => {
	const singleResDatum = await dataSource.getSingle(mainResourceId);

	if (singleResDatum) {
		const theFormatted = serializerPair.serialize(singleResDatum);
		res.writeHead(constants.HTTP_STATUS_OK, { 'Content-Type': mediaType });
		res.end(theFormatted);
		return;
	}

	res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
	res.statusMessage = `${resource.itemName} Not Found`;
	res.end();
	return;
};

const handleCreate = async (
	deserializer: SerializerPair,
	serializer: SerializerPair,
	mediaType: string,
	resource: Resource,
	dataSource: DataSource,
	req: IncomingMessage,
	res: ServerResponse
) => {
	return new Promise<void>((resolve) => {
		let body = Buffer.from('');
		req.on('data', (chunk) => {
			body = Buffer.concat([body, chunk]);
		});
		req.on('end', async () => {
			const bodyStr = body.toString('utf-8'); // TODO use encoding in request header
			let bodyDeserialized: object;
			try {
				bodyDeserialized = deserializer.deserialize(bodyStr);
				if (typeof bodyDeserialized !== 'object' || bodyDeserialized === null) {
					res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
					res.statusMessage = `Invalid ${resource.itemName}`;
					res.end();
					resolve();
					return;
				}
			} catch {
				res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
				res.statusMessage = `Invalid ${resource.itemName}`;
				res.end();
				resolve();
				return;
			}

			try {
				const newId = await resource.newId(dataSource);
				const newObject = await dataSource.create({
					...bodyDeserialized,
					[resource.idAttr]: newId,
				});
				const theFormatted = serializer.serialize(newObject);
				res.writeHead(constants.HTTP_STATUS_OK, {'Content-Type': mediaType});
				res.end(theFormatted);
				return;
			} catch {
				res.statusCode = constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE;
				res.statusMessage = `Could Not Return ${resource.itemName}`;
				res.end();
			}

			resolve();
		});
	});
};

export const application = (appParams: ApplicationParams) => {
	const resources = new Set<Resource>();
	const serializers = new Map<string, SerializerPair>();

	return {
		contentType(mimeTypePrefix: string, serializerPair: SerializerPair) {
			serializers.set(mimeTypePrefix, serializerPair);
			return this;
		},
		resource(res: Partial<Resource>) {
			res.dataSource = res.dataSource ?? appParams.dataSource?.(res as Resource);
			if (typeof res.dataSource === 'undefined') {
				throw new Error(`Resource ${res.itemName} must have a data source.`);
			}
			resources.add(res as Resource);
			return this;
		},
		async createServer(serverParams = {} as CreateServerParams) {
			const {
				baseUrl = '/',
				host = 'http://localhost' // TODO not a sensible default...
			} = serverParams;

			const serverModule = await import('http');
			return serverModule.createServer(
				async (req, res) => {
					if (!req.method) {
						res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
							'Allow': 'HEAD,GET,POST,PUT,PATCH,DELETE' // TODO check with resources on allowed methods
						});
						res.end();
						return;
					}

					if (!req.url) {
						res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
						res.end();
						return;
					}

					const urlObject = new URL(req.url, host);
					const urlWithoutBaseRaw = urlObject.pathname.slice(baseUrl.length);
					const urlWithoutBase = urlWithoutBaseRaw.length < 1 ? '/' : urlWithoutBaseRaw;

					if (req.method.toUpperCase() === 'GET' && urlWithoutBase === '/') {
						const data = {
							name: appParams.name
						};
						res.writeHead(constants.HTTP_STATUS_OK, {
							'Content-Type': 'application/json', // TODO content negotiation,
							// we are using custom headers for links because the standard Link header
							// is referring to the document metadata (e.g. author, next page, etc)
							// https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Link
							'X-Resource-Link': Array.from(resources)
								.map((r) =>
										`<${baseUrl}/${r.routeName}>; name="${r.collectionName}"`,
									// TODO add host?
								)
								.join(', ')
						});
						res.end(JSON.stringify(data))
						return;
					}

					const [, mainResourceRouteName, mainResourceId = ''] = urlWithoutBase.split('/');
					const theResource = Array.from(resources).find((r) => r.routeName === mainResourceRouteName);
					if (typeof theResource !== 'undefined') {
						await theResource.dataSource.initialize();
						const method = req.method.toUpperCase();
						if (method === 'GET') {
							const negotiator = new Negotiator(req);
							const availableMediaTypes = Array.from(serializers.keys());
							const theMediaType = negotiator.mediaType(availableMediaTypes);

							if (typeof theMediaType === 'undefined') {
								res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
								res.end();
								return;
							}

							const theSerializerPair = serializers.get(theMediaType);
							if (typeof theSerializerPair === 'undefined') {
								res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
								res.end();
								return;
							}

							if (mainResourceId === '') {
								await handleGetAll(theSerializerPair, theMediaType, theResource.dataSource, res);
								return;
							}

							await handleGetSingle(theSerializerPair, theMediaType, theResource, mainResourceId, theResource.dataSource, res);
							return;
						}

						if (method === 'POST') {
							if (mainResourceId === '') {
								const theDeserializer = serializers.get(req.headers['content-type'] ?? 'application/octet-stream');
								if (typeof theDeserializer === 'undefined') {
									res.statusCode = constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE;
									res.end();
									return;
								}

								const negotiator = new Negotiator(req);
								const availableMediaTypes = Array.from(serializers.keys());
								const theMediaType = negotiator.mediaType(availableMediaTypes);

								if (typeof theMediaType === 'undefined') {
									res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
									res.end();
									return;
								}

								const theSerializer = serializers.get(theMediaType);
								if (typeof theSerializer === 'undefined') {
									res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
									res.end();
									return;
								}

								await handleCreate(theDeserializer, theSerializer, theMediaType, theResource, theResource.dataSource, req, res);
								return;
							}

							res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
							res.end();
							return;
						}

						return;
					}

					res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
					res.statusMessage = 'URL Not Found';
					res.end();
				}
			);
		}
	};
};
